import 'dart:async';
import 'dart:typed_data';
import 'package:flutter/services.dart';
import 'package:ffi/ffi.dart';

import 'dart:ffi';
import 'dart:io';

//How C Function looks like and how it's going to be exposed to dart
typedef flutter_oboe_native_add = Int32 Function(Int32 a, Int32 b); //Typedef for C
typedef FlutterOboeNativeAdd = int Function(int a, int b); //Typedef for dart

//Oboe Type defs
//Initialisations
typedef oboe_engine_create = Pointer<Void> Function();
typedef OboeEngineCreate = Pointer<Void> Function();

typedef oboe_engine_dispose = Void Function(Pointer<Void>);
typedef OboeEngineDispose= void Function(Pointer<Void>);

//Recorder
typedef oboe_engine_start_recording = Void Function(Pointer<Void>);
typedef OboeEngineStartRecording = void Function(Pointer<Void>);

typedef oboe_engine_stop_recording = Void Function(Pointer<Void>);
typedef OboeEngineStopRecording = void Function(Pointer<Void>);

typedef oboe_engine_write_to_file = Void Function(Pointer<Void>, Pointer<Utf8>);
typedef OboeEngineWriteToFile = void Function(Pointer<Void>, Pointer<Utf8>);

//Player
typedef oboe_engine_play_from_file = Void Function(Pointer<Void>, Pointer<Utf8>);
typedef OboeEnginePlayFromFile = void Function(Pointer<Void>, Pointer<Utf8>);

typedef oboe_engine_stop_playing_from_file = Void Function(Pointer<Void>);
typedef OboeEngineStopPlayingFromFile = void Function(Pointer<Void>);

class OboeAudioEngine {
  final nativeLib = Platform.isAndroid ? DynamicLibrary.open("liboboeffi.so") : DynamicLibrary.process();

  //Singleton class so that it can be called with FlutterObe()
  static final OboeAudioEngine _instance =  OboeAudioEngine._();
  factory OboeAudioEngine(){
    // if(_instance == null) {
    //   _instance = OboeAudioEngine._();
    // }
    return _instance;
  }

  //Audio stream
  OboeEngineCreate? _engineCreate;
  OboeEngineDispose? _engineDispose;

  //Audio recording
  OboeEngineStartRecording? _engineStartRecording;
  OboeEngineStopRecording? _engineStopRecording;
  OboeEngineWriteToFile? _engineWriteToFile;

  //Audio player
  OboeEnginePlayFromFile? _enginePlayFromFile;
  OboeEngineStopPlayingFromFile? _engineStopPlayingFromFile;

  //Android build the native code as a shared object(<lib_name>.so)
  //XCode toolchain complile the native code directly into the main executable and can just call process method
  OboeAudioEngine._(){
    _engineCreate = nativeLib.lookup<NativeFunction<oboe_engine_create>>('oboe_create').asFunction();
    _engineDispose = nativeLib.lookup<NativeFunction<oboe_engine_dispose>>('oboe_dispose').asFunction();

    //Audio recorder
    _engineStartRecording = nativeLib.lookup<NativeFunction<oboe_engine_start_recording>>('start_recording').asFunction();
    _engineStopRecording = nativeLib.lookup<NativeFunction<oboe_engine_stop_recording>>('stop_recording').asFunction();
    _engineWriteToFile = nativeLib.lookup<NativeFunction<oboe_engine_write_to_file>>('write_to_file').asFunction();

    //Audio player
    _enginePlayFromFile = nativeLib.lookup<NativeFunction<oboe_engine_play_from_file>>('play_from_file').asFunction();
    _engineStopPlayingFromFile = nativeLib.lookup<NativeFunction<oboe_engine_stop_playing_from_file>>('stop_playing_from_file').asFunction();
  }
}

//Wrapper to access from public
class FlutterOboe{
  late Pointer<Void>? _nativeInstance;
  FlutterOboe(){
    _nativeInstance = OboeAudioEngine()._engineCreate!();
  }

  void dispose(){
    OboeAudioEngine()._engineDispose!(_nativeInstance!);
  }

  //Audio Recorder
  void startRecording(){
    OboeAudioEngine()._engineStartRecording!(_nativeInstance!);
  }

  void stopRecording(){
    OboeAudioEngine()._engineStopRecording!(_nativeInstance!);
  }

  void writeToFile(String path)
  {
    OboeAudioEngine()._engineWriteToFile!(_nativeInstance!,  path.toNativeUtf8());
  }

  void playFromFile(String path)
  {
    OboeAudioEngine()._enginePlayFromFile!(_nativeInstance!,  path.toNativeUtf8());
  }

  void stopPlayingFromFile(){
    OboeAudioEngine()._engineStopPlayingFromFile!(_nativeInstance!);
  }
}
